/*
 * Software for Interactive Project Pool Game
 * Columbia University. New York, 2008
 * Authors:
 *   Abdulhamid Ghandour
 *   Thomas John
 *   Jaime Peretzman
 *   Bharadwaj Vellore
 * 
 * Desc:
 */
#include <stdio.h>
#include "debug.h"
#include "fixedpoint.h"
#include "gameconfig.h"
#include "types.h"
#include "ball.h"
#include "i2c.h"
#include "ui.h"
#include "calibration.h"

#define NO_CUE_DETECTED (IORD_32DIRECT(VISION_BASE, 0) >> 31)

static void initPockets();
static void initBalls();
void initPlayers();
static void initCue();
static int play();
extern int calibrate();
struct ball_t balls[NUM_BALLS];
struct ball_t cue;
struct ball_t pockets[6];
struct player_t player1, player2;
struct player_t *pCurrentPlayer;

typedef enum gameState_e {
  GAME_PLAYING,
  GAME_WAITING_TO_PLAY
}gameState_t;
gameState_t gameState = GAME_WAITING_TO_PLAY;

void showPoints(struct player_t *player);

int main(){
  DP_HI;

  DP("Welcome to Projection Pool\n");

  LCD_Init();
  initPointsDisplay();

  LCD_Show_Text("Welcome to Pool!");

  initPlayers();
  configureCamera();
  doCalibration();
  CALIBRATED;
  initCue();
  initPockets();
  initBalls();
  initPointsDisplay();
  drawBalls(balls);
  STARTED_GAME;
    
  while(1){
    if(-1 == play()){
      doCalibration();
      CALIBRATED;
      drawBalls(balls);     
    }else{
      initPlayers();
      initCue();
      initBalls();
      initPointsDisplay();
      drawBalls(balls);
      STARTED_GAME;
    }
  }

  DP_BYE;
}

long long calibScaleHoz, calibOffsetHoz;
long long calibScaleVer, calibOffsetVer;
void configureScaling(){
  long long vgaHozRes = 640;
  long long vgaVerRes = 480;
  long long blackMarginHoz = BLACK_MARGIN_HOZ;
  long long blackMarginVer = BLACK_MARGIN_VER;

  calibScaleHoz = FPDIV(
    INT2FP(vgaHozRes - 2*blackMarginHoz),
    INT2FP(camEndX - camStartX)
  );
  calibOffsetHoz = 
    FPMUL(calibScaleHoz,INT2FP(CAMERA_CROP_MARGIN_HOZ)) + 
    INT2FP(blackMarginHoz);

  calibScaleVer = FPDIV(
    INT2FP(vgaVerRes - 2*blackMarginVer),
    INT2FP(camEndY - camStartY)
  );
  calibOffsetVer = 
    FPMUL(calibScaleVer,INT2FP(CAMERA_CROP_MARGIN_VER)) + 
    INT2FP(blackMarginVer);
}

int play(){
  int i,j;
  long long time, earliestEventTime, eventTime;
  event_t eventType;
  struct ball_t *collidingBall1, *collidingBall2;
  edge_t tableEdge, collisionTableEdge;
  bool_t ballsMoved = FALSE;
  bool_t repositionedWhite;

  long long xin = 0, prevXin = 0;
  long long yin = 0, prevYin = 0;
  long long xinBeforeFilter = 0, yinBeforeFilter = 0;
  long long A = FPDIV(INT2FP(18),INT2FP(10));
  long long W0, W1, W2, W3;

  W0 = FPDIV(INT2FP(5),INT2FP(10)); 
  W1 = FPDIV(INT2FP(3),INT2FP(10));
  W2 = FPDIV(INT2FP(1),INT2FP(10));
  W3 = FPDIV(INT2FP(1),INT2FP(10));
  
  struct vector prevCuePos1, prevCuePos2, prevCuePos3;
  prevCuePos1.x = 0;
  prevCuePos1.y = 0;
  prevCuePos2.x = 0;
  prevCuePos2.y = 0;
  prevCuePos3.x = 0;
  prevCuePos3.y = 0;
  
  long long prevXinBeforeFilter = 0, prevYinBeforeFilter = 0;
  long long absPreFilterXDiff, absPreFilterYDiff;

  int numBallsPocketed = 0;
  gameState = GAME_WAITING_TO_PLAY;

  configureScaling();
  LCD_Init();
  LCD_Show_Text("Player");
  LCD_Line2();
  LCD_Show_Text("1");

  while(1){
    time = INT2FP(1LL);
    if(!NO_CUE_DETECTED){
      xin = (long long)IORD_32DIRECT(VISION_BASE, 0);
      yin = ((xin >> 16) & 0x7FFF);
      xin = (xin & 0xFFFF);
      xinBeforeFilter = FPMUL(calibScaleHoz,INT2FP(xin)) + calibOffsetHoz;
      yinBeforeFilter = FPMUL(calibScaleVer,INT2FP(yin)) + calibOffsetVer;

      /*
       * Filter to limit step changes in cue position
       */
      absPreFilterXDiff = (xinBeforeFilter > prevXinBeforeFilter)?
                          (xinBeforeFilter - prevXinBeforeFilter):
                          (prevXinBeforeFilter > xinBeforeFilter);
      absPreFilterYDiff = (yinBeforeFilter > prevYinBeforeFilter)?
                          (yinBeforeFilter - prevYinBeforeFilter):
                          (prevYinBeforeFilter > yinBeforeFilter);

      if((prevXinBeforeFilter != 0) && (prevXinBeforeFilter !=0)){
        if((absPreFilterXDiff > INT2FP(75)) || (absPreFilterYDiff > INT2FP(75))){
          xinBeforeFilter = prevXinBeforeFilter;
          yinBeforeFilter = prevYinBeforeFilter;  
        }else{
          prevXinBeforeFilter = xinBeforeFilter;
          prevYinBeforeFilter = yinBeforeFilter; 
        }
      }else{
        prevXinBeforeFilter = xinBeforeFilter;
        prevYinBeforeFilter = yinBeforeFilter;
      }

      /*
       * Smoothing filter for cue position
       */
      xin = FPMUL(W0, xinBeforeFilter) + 
            FPMUL(W1, prevCuePos1.x) + 
            FPMUL(W2, prevCuePos2.x) + 
            FPMUL(W3, prevCuePos3.x);
      yin = FPMUL(W0, yinBeforeFilter) + 
            FPMUL(W1, prevCuePos1.y) + 
            FPMUL(W2, prevCuePos2.y) + 
            FPMUL(W3, prevCuePos3.y);
      
      prevCuePos3.x = prevCuePos2.x;
      prevCuePos3.y = prevCuePos2.y;
      prevCuePos2.x = prevCuePos1.x;
      prevCuePos2.y = prevCuePos1.y;
      prevCuePos1.x = xinBeforeFilter;
      prevCuePos1.y = yinBeforeFilter;

      cue.ballState = BALL_VISIBLE;

      if ((prevXin != 0) && (prevYin != 0)){
        cue.pos.x = prevXin;
        cue.pos.y = prevYin;
        if(IORD_32DIRECT(UICONTROL_BASE, 10 * 4) & 0x00000400){
            IOWR_POS(VGA_BASE, 13, FP2INT(prevXin));
            IOWR_POS(VGA_BASE, 14, FP2INT(prevYin));
        }
        else{
            IOWR_POS(VGA_BASE, 13, 999);
            IOWR_POS(VGA_BASE, 14, 999);
        }
        cue.vel.x = FPMUL(A,(xin - prevXin));
        cue.vel.y = FPMUL(A,(yin - prevYin));
        prevXin = xin;
        prevYin = yin;
      }
      else{
        prevXin = xin;
        prevYin = yin;
      }
    }else{
      cue.ballState = BALL_INVISIBLE;
      prevXin = 0;
      prevYin = 0;
      prevXinBeforeFilter = 0;
      prevYinBeforeFilter = 0;  
    }

    ballsMoved = FALSE;
    do{
      earliestEventTime = time;
      eventType = NO_COLLISION;
      collidingBall1 = NULL;
      collidingBall2 = NULL;
      tableEdge = NONE;

      for(i=0; i<NUM_BALLS; i++){
        if(BALL_INVISIBLE == balls[i].ballState){
          continue;
        }
        if(isBallMoving(&balls[i])){
          ballsMoved = TRUE;
        }

        /*
        // Check for collisions with the table boundaries
        */
        if(isBallMoving(&balls[i]) == TRUE){
          eventTime = collisionWithTableTime(&balls[i],&tableEdge);
          if((eventTime >= 0) && (eventTime < earliestEventTime)){
            earliestEventTime = eventTime;
            collidingBall1 = &balls[i];
            collidingBall2 = NULL;
            eventType = TABLE_COLLISION;
            collisionTableEdge = tableEdge;
          }
        }

        /*
        // Check for "collision" with cue
        */
        if(BALL_INVISIBLE != cue.ballState){
          if((balls[i].colour == COL_WHITE) ||
              ((IORD_32DIRECT(UICONTROL_BASE, 10 * 4) & 0x1000) == 0)
          ){            
            if(((balls[i].vel.x == 0) && (balls[i].vel.y == 0)) ||
               ((IORD_32DIRECT(UICONTROL_BASE, 10 * 4) & 0x800) == 0)
            ){
              eventTime = collisionWithBallTime(&balls[i],&cue);
              if((eventTime >= 0) && (eventTime < earliestEventTime)){
                earliestEventTime = eventTime;
                collidingBall1 = &balls[i];
                collidingBall2 = &cue;
                eventType = CUE_COLLISION;
              }
            }
          }
        }

        /*
        // Check for "collision" with pockets
        */
        if(isBallMoving(&balls[i]) == TRUE){
          for(j=0; j<6; j++){      
            eventTime = collisionWithBallTime(&balls[i], &pockets[j]);
            if((eventTime >= 0) && (eventTime < earliestEventTime)){
              earliestEventTime = eventTime;
              collidingBall1 = &balls[i];
              collidingBall2 = &pockets[j];
              eventType = POCKET_COLLISION;
            }
          }
        }

        /*
        // Collision with other balls
        */
        if(isBallMoving(&balls[i]) == TRUE){
          for(j=0; j<NUM_BALLS; j++){
            if(BALL_INVISIBLE == balls[j].ballState){
              continue;
            }

            eventTime = collisionWithBallTime(&balls[i],&balls[j]);
            if((eventTime >= 0) && (eventTime < earliestEventTime)){
              earliestEventTime = eventTime;
              collidingBall1 = &balls[i];
              collidingBall2 = &balls[j];
              eventType = BALL_COLLISION;
            }
          }
        }
      }

      moveBalls(balls, earliestEventTime);

      switch(eventType){
        case NO_COLLISION:
        break;

        case POCKET_COLLISION:
          collidingBall1->ballState = BALL_INVISIBLE;
          numBallsPocketed++;
          PLAY_SOUND;
          pCurrentPlayer->points += collidingBall1->points;
          showPoints(pCurrentPlayer);

          if(collidingBall1->colour == COL_WHITE){
            int k;
            repositionedWhite = TRUE;
            numBallsPocketed--;
            collidingBall1->ballState = BALL_VISIBLE;
            collidingBall1->pos.x = TABLE_END_X - INT2FP(140);
            collidingBall1->pos.y = TABLE_END_Y - INT2FP(120);
            printf("New Pos = %lld, %lld\n",
              FP2INT(collidingBall1->pos.x),
              FP2INT(collidingBall1->pos.y)
            );
            collidingBall1->vel.x = 0;
            collidingBall1->vel.y = 0;
            while(1){
              long long absXDiff, absYDiff;
              for(k=0; k<NUM_BALLS; k++){
                if(&balls[k] == collidingBall1){
                  continue;
                }
                absXDiff = (collidingBall1->pos.x > balls[k].pos.x)?
                           (collidingBall1->pos.x - balls[k].pos.x):
                           (balls[k].pos.x - collidingBall1->pos.x);
                absYDiff = (collidingBall1->pos.y > balls[k].pos.y)?
                           (collidingBall1->pos.y - balls[k].pos.y):
                           (balls[k].pos.y - collidingBall1->pos.y);
                if((absXDiff < INT2FP(30)) && (absYDiff < INT2FP(30))){
                  repositionedWhite = FALSE;
                  break;                    
                }
              }
              if(repositionedWhite == FALSE){
                collidingBall1->pos.x += INT2FP(15);
                collidingBall1->pos.y += INT2FP(10);
                if(collidingBall1->pos.x > (TABLE_END_X - INT2FP(20))){
                  collidingBall1->pos.x -= INT2FP(200);
                }
                if(collidingBall1->pos.y > (TABLE_END_Y - INT2FP(20))){
                  collidingBall1->pos.y -= INT2FP(150);
                }
                printf("New Pos = %lld, %lld\n",
                  FP2INT(collidingBall1->pos.x),
                  FP2INT(collidingBall1->pos.y)
                );
                repositionedWhite = TRUE;
              }else{
                break;
              }
            }
          }
        break;

        case CUE_COLLISION:
          handleCollisionWithCue(collidingBall1,&cue);
          PLAY_SOUND;
        break;

        case TABLE_COLLISION:
          switch(collisionTableEdge){
            case LEFT:
            case RIGHT:
              collidingBall1->vel.x *= -1LL;
            break;

            case TOP:
            case BOTTOM:
              collidingBall1->vel.y *= -1LL;
            break;

            default:
              DP_ASSERT(0,"Collision with non-existent table edge!");
            break;
          }
          PLAY_SOUND;        
        break;

        case BALL_COLLISION:
          handleBallCollision(collidingBall1, collidingBall2);
          PLAY_SOUND;
        break;

        default:
          DP_ASSERT(0,"Invalid event");
        break;
      };
      drawBalls(balls);
      time -= earliestEventTime;
    }while(time > 0);
    drawBalls(balls);
    applyFriction(balls);

    /*
     * Check if there have been balls moved in the last time step
     * If yes, and if the previous state was WAITING_TO_PLAY, then
     * switch to PLAYING_STATE. Continue with the same player.
     * If not, and if the previous state was PLAYING, then switch
     * to the WAITING_TO_PLAY. Also switch players.
     */   
    if((GAME_PLAYING == gameState) && (FALSE == ballsMoved)){
      gameState = GAME_WAITING_TO_PLAY;
      if(pCurrentPlayer == &player1){
        pCurrentPlayer = &player2;
        LCD_Line2();
        LCD_Show_Text("2");
      }else if(pCurrentPlayer == &player2){
        pCurrentPlayer = &player1;
        LCD_Line2();
        LCD_Show_Text("1");
      }else{
        fflush(stdout);
        DP_ASSERT(0,"Invalid player\n");
      }
    }else if((GAME_WAITING_TO_PLAY == gameState) && (TRUE == ballsMoved)){
      gameState = GAME_PLAYING;
    }else{
      /*
       * Do nothing
       */    
    }

    /*
     * Check if there has been a request for re-calibration
     * and return -1 if yes
     */
    if(CALIBRATION_REQUESTED == 0){
      if(numBallsPocketed == NUM_BALLS - 1){
        initBalls();
        initPlayers();
        initCue();
        initPointsDisplay();
      }
      return -1;
    }
    
    if(NEW_GAME_REQUESTED == 0){
      return 0;
    }

    if(numBallsPocketed == NUM_BALLS - 1){
      LCD_Init();
      LCD_Show_Text("Player");
      if(player1.points > player2.points){
        LCD_Show_Text(" 1 WINS!");        
      }else if(player1.points < player2.points){
        LCD_Show_Text(" 2 WINS!");
      }else{
        LCD_Show_Text("s TIE!");
      }
      while(NEW_GAME_REQUESTED == 1);
      return 0;
    }
  }
}

static void initPockets(){
  int i;

  for(i=0; i<6; i++){
    pockets[i].vel.x = 0;
    pockets[i].vel.y = 0;
    pockets[i].radius = POCKET_RADIUS;
    pockets[i].colour = 1;
    pockets[i].ballState = BALL_INVISIBLE;
  }

  pockets[0].pos.x = TOP_LEFT_POCKET_X;
  pockets[0].pos.y = TOP_LEFT_POCKET_Y; 
  pockets[1].pos.x = TOP_MID_POCKET_X;
  pockets[1].pos.y = TOP_MID_POCKET_Y; 
  pockets[2].pos.x = TOP_RIGHT_POCKET_X;
  pockets[2].pos.y = TOP_RIGHT_POCKET_Y; 
  pockets[3].pos.x = BOTTOM_LEFT_POCKET_X;
  pockets[3].pos.y = BOTTOM_LEFT_POCKET_Y; 
  pockets[4].pos.x = BOTTOM_MID_POCKET_X;
  pockets[4].pos.y = BOTTOM_MID_POCKET_Y; 
  pockets[5].pos.x = BOTTOM_RIGHT_POCKET_X;
  pockets[5].pos.y = BOTTOM_RIGHT_POCKET_Y;
}

static void initBalls(){
  int i;

  for(i=0; i<NUM_BALLS; i++){
    balls[i].radius = BALL_RADIUS;
    balls[i].ballState = BALL_VISIBLE;
  }
  balls[0].colour = COL_WHITE;
  balls[0].points = -10;

  balls[1].colour = COL_YELLOW;
  balls[1].points = 20;

  balls[2].colour = COL_CYAN;
  balls[2].points = 5;

  balls[3].colour = COL_K3;
  balls[3].points = 10;

  balls[4].colour = COL_K3;
  balls[4].points = 10;

  balls[5].colour = COL_CYAN;
  balls[5].points = 5;

  balls[6].colour = COL_CYAN;
  balls[6].points = 5;
  
  balls[0].pos.x = INT2FP(tableStartX + 240);
  balls[0].pos.y = INT2FP(tableStartY + 120);
  balls[0].vel.x = FPDIV(INT2FP(0LL),INT2FP(1LL));
  balls[0].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));

  balls[1].pos.x = INT2FP(tableStartX + 40);
  balls[1].pos.y = INT2FP(tableStartY + 120);
  balls[1].vel.x = FPDIV(INT2FP(0LL),INT2FP(2LL));
  balls[1].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));

  balls[2].pos.x = INT2FP(tableStartX + 40);
  balls[2].pos.y = INT2FP(tableStartY + 160);
  balls[2].vel.x = FPDIV(INT2FP(0LL),INT2FP(1LL));
  balls[2].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));

  balls[3].pos.x = INT2FP(tableStartX + 100);
  balls[3].pos.y = INT2FP(tableStartY + 100);
  balls[3].vel.x = FPDIV(INT2FP(0LL),INT2FP(1LL));
  balls[3].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));

  balls[4].pos.x = INT2FP(tableStartX + 100);
  balls[4].pos.y = INT2FP(tableStartY + 140);
  balls[4].vel.x = FPDIV(INT2FP(0LL),INT2FP(1LL));
  balls[4].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));

  balls[5].pos.x = INT2FP(tableStartX + 150);
  balls[5].pos.y = INT2FP(tableStartY + 120);
  balls[5].vel.x = FPDIV(INT2FP(0LL),INT2FP(1LL));
  balls[5].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));
  
  balls[6].pos.x = INT2FP(tableStartX + 40);
  balls[6].pos.y = INT2FP(tableStartY + 80);
  balls[6].vel.x = FPDIV(INT2FP(0LL),INT2FP(1LL));
  balls[6].vel.y = FPDIV(INT2FP(0LL),INT2FP(1LL));
}

static void initCue(){
  cue.colour = 0LL;
  cue.pos.x = INT2FP(10LL);
  cue.pos.y = INT2FP(10LL);
  cue.radius = INT2FP(2LL);
  cue.vel.x = 0LL;
  cue.vel.y = 0LL;
}

void initPlayers(){
   player1.points = 0;
   player2.points = 0;
   pCurrentPlayer = &player1;
}

void showPoints(struct player_t *player){
  if(player == &player1){
    if(player->points < 0){
      HEX5(sevensegment[0]);
      HEX4(sevensegment[0]);
    }else{
      HEX5(sevensegment[player->points/10]);
      HEX4(sevensegment[player->points%10]);
    }
  }else if(player == &player2){
    if(player->points < 0){
      HEX1(sevensegment[0]);
      HEX0(sevensegment[0]);
    }else{
      HEX1(sevensegment[player->points/10]);
      HEX0(sevensegment[player->points%10]);
    }
  }else{
    DP_ASSERT(0,"Invalid player"); 
  }
}
